
/* written by Adrian Schroeter <adrian@suse.de> based on the sync-fstab code from hal *
 * licensed under GPL                                                                 */

#define _GNU_SOURCE
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <fstab.h>
#include <stdio.h>
#include <resmgr.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/file.h>
#include <time.h>
#include <unistd.h>
#include <mntent.h>
#include <syslog.h>

#include <popt.h>

#define DBUS_API_SUBJECT_TO_CHANGE
#include "hal/libhal.h"
#include "hal/libhal-storage.h"

#define MOUNT_LENGTH 2048    
#define FALSE 0
#define TRUE  1

static LibHalContext *hal_context = NULL;

static int     fsy_use_managed;
static char   *fsy_mount_root;
static char   *fsy_managed_primary;
static char   *fsy_managed_secondary;

static int debug = FALSE;
static pid_t pid;

static const char char_white_list[] = "_-+abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";

int
main (int argc, const char *argv[])
{
  int i;
  poptContext popt_context;
  char *udi_to_add = NULL, *udi_to_remove = NULL, *hal_device_udi, *caps;
  const char **left_over_args = NULL;

  pid = getpid ();

  openlog ("hal-subfs-mount", LOG_PID, LOG_USER);

  struct poptOption options[] = {
      {"add",     'a', POPT_ARG_STRING, &udi_to_add,    0, "add an entry to fstab", "UDI"},
      {"remove",  'r', POPT_ARG_STRING, &udi_to_remove, 0, "remove an entry from fstab", "UDI"},

      POPT_AUTOHELP

      {NULL, '\0', 0, NULL, 0, NULL, NULL}
  };

  popt_context = poptGetContext ("hal-subfs-mount", argc, argv, options, 0);

  while ((i = poptGetNextOpt (popt_context)) != -1)
    {
      if (i < -1)
        {
          poptPrintHelp (popt_context, stderr, 0);
          return 1;
        }
    }

  if (getenv ("HALD_VERBOSE") != NULL || getenv ("SUBFS_DEBUG") != NULL)
      debug = TRUE;
//debug=TRUE;
  if (debug)
      syslog (LOG_INFO, "hal-subfs-mount started in debug mode. " );

  /* accept "add" and "remove" / HAL environment variables
   * in addition to "--add" and "--remove" above
   */
  left_over_args = poptGetArgs (popt_context);

  hal_device_udi = getenv ("UDI");
  if ( !hal_device_udi ){
  	if (debug)
      		syslog (LOG_INFO, "no UDI in enviroment " );
	return 0;
  }
  
  caps = getenv ("HAL_PROP_INFO_CAPABILITIES");
  if (caps == NULL) {
  	if (debug)
      		syslog (LOG_INFO, "HAL_PROP_INFO_CAPABILITIES is set" );
	return 0;
  }

  if (strstr (caps, "volume") == NULL &&
      strstr (caps, "storage") == NULL) {
  	if (debug)
      		syslog (LOG_INFO, "neither a voume nor a storage device" );
	return 0;
  }

  // no actions on hald restart
  if (getenv ("HALD_SHUTDOWN") != NULL) {
  	if (debug)
      		syslog (LOG_INFO, "hald is in shutdown mode" );
	return 0;
  }

  for (i = 0; left_over_args[i] != NULL; i++)
  {
          if (strcmp (left_over_args[i], "add") == 0 && udi_to_add == NULL)
          {
                  udi_to_add = strdup (hal_device_udi);
                  break;
          }

          if (strcmp (left_over_args[i], "remove") == 0)
          {
                  udi_to_remove = strdup (hal_device_udi);
                  break;
          }
  }

  if ( !udi_to_remove && !udi_to_add ) {
  	if (debug)
      		syslog (LOG_INFO, "neither an add nor remove event" );
	return 0;
  }

  if (debug)
      syslog (LOG_INFO, "hal-subfs-mount init hal lib " );
  // initialize hal lib
  hal_context = hal_initialize (NULL, FALSE);
  fsy_mount_root = hal_drive_policy_default_get_mount_root (hal_context);
  if (fsy_mount_root == NULL) {
  	  if (debug)
      		syslog (LOG_INFO, "empty hal_drive_policy_default_get_mount_root " );
          return 0;
  }
  fsy_use_managed = hal_drive_policy_default_use_managed_keyword (hal_context);
  if (fsy_use_managed) {
          if ((fsy_managed_primary   = hal_drive_policy_default_get_managed_keyword_primary (hal_context)) == NULL) {
	          return 0;
	  }
          if ((fsy_managed_secondary = hal_drive_policy_default_get_managed_keyword_secondary (hal_context)) == NULL) {
	          return 0;
	  }
  }

  if (debug)
      syslog (LOG_INFO, "hal-subfs-mount  hal lib initialized " );

  HalDrive *drive;
  HalVolume *volume;
  const char *device_file;

  drive = hal_drive_from_udi (hal_context, hal_device_udi);
  if (drive == NULL) {
	char *udi_storage_device;
	/* try block.storage_device */
	udi_storage_device = hal_device_get_property_string (hal_context, hal_device_udi, "block.storage_device");
	if (udi_storage_device == NULL)
	        return -1;
	drive = hal_drive_from_udi (hal_context, udi_storage_device);
	hal_free_string (udi_storage_device);
	if (drive == NULL)
	        return -1;
  }
  if (debug)
      syslog (LOG_INFO, "hal-subfs-mount got drive" );

  
  volume = hal_volume_from_udi (hal_context, hal_device_udi);
  // avoid double mounts during hald startup
  if ( !volume && hal_drive_get_type(drive) == HAL_DRIVE_TYPE_CDROM ){
     if (debug)
        syslog (LOG_INFO, "hal-subfs-mount is_disc, but no volume" );
     return 1;
  }
  if ( volume && !hal_volume_policy_is_mountable (drive, volume, NULL)){
     if (debug)
        syslog (LOG_INFO, "hal-subfs-mount not mountable volume" );
     return 1;
  }
  if ( !volume && !hal_drive_policy_is_mountable (drive, NULL)){
     if (debug)
        syslog (LOG_INFO, "hal-subfs-mount not mountable drive" );
     return 1;
  }

  if ( volume )
    device_file = hal_volume_get_device_file (volume);
  else
    device_file = hal_drive_get_device_file (drive);
  if ( !device_file ){
     if (debug)
        syslog (LOG_INFO, "hal-subfs-mount no device node" );
     return 1;
  }

  // resmgr handling
  if (udi_to_add) {
    // register at resmgr
    rsm_command(NULL, "add %s desktop scsi", device_file );
    if (debug)
        syslog (LOG_INFO, "hal-subfs-mount %s registered at resmgr", device_file );
  }
  if (udi_to_remove){
    rsm_remove_device( device_file, "desktop" );
    if (debug)
        syslog (LOG_INFO, "hal-subfs-mount %s removed at resmgr", device_file );
  }

  char mount_point[MOUNT_LENGTH+1];
  const char *desired_mount_point;
  struct mntent *fe;
  int counter = 0;

  if ( volume ){
     if ( !(desired_mount_point = hal_volume_policy_get_desired_mount_point (drive, volume, NULL)) ){
        if (debug)
           syslog (LOG_INFO, "hal-subfs-mount no desired mount point" );
        desired_mount_point = strdup( "hal_volume" );
     }
  }else{
     if ( !(desired_mount_point = hal_drive_policy_get_desired_mount_point (drive, NULL)) ){
        if (debug)
           syslog (LOG_INFO, "hal-subfs-mount no desired mount point" );
        desired_mount_point = strdup( "hal_volume" );
     }
  }
  snprintf( mount_point, MOUNT_LENGTH, "/media/%s", desired_mount_point);

  while ( counter < 1024 ){
    counter++;

    // allow only these chars in mount point
    char *p, *c;
    p = mount_point + strlen("/media/");

    for ( ; *p; p++ ){
       for ( c = char_white_list; *c; c++ )
	  if ( *c == *p )
	     break;

       if ( !*c )
	  *p = '_';
    }

    FILE *f = setmntent( "/etc/mtab", "r" );
    if ( !f ){
       if (debug)
          syslog (LOG_INFO, "ERROR: unable to parse /etc/mtab" );
       return 1;
    }
    while ( (fe = getmntent(f)) && fe && strcmp(fe->mnt_dir, mount_point));
    endmntent( f );

    if ( !fe ){
       if (debug)
          syslog (LOG_INFO, "free mountpoint %s ", mount_point );
       break;
    };

    snprintf( mount_point, MOUNT_LENGTH, "/media/%s_%d", desired_mount_point, counter);
  }

  // mount device
  if (udi_to_add) {
    char subfs_type[MOUNT_LENGTH+1];
    char cmd[MOUNT_LENGTH+1];
//    char *fstype, *options;

    if (debug){
       syslog (LOG_INFO, "DRIVE %s TYPE %d.",
               device_file, hal_drive_get_type(drive) );
    }
    if ( hal_drive_get_type(drive) == HAL_DRIVE_TYPE_CDROM )
	strcpy( subfs_type, "fs=cdfss,ro,procuid," );
    else
	strcpy( subfs_type, "fs=floppyfss,sync,procuid," );

    if ( volume ){
     char *fstype;
     if ( (fstype = hal_volume_policy_get_mount_fs (drive, volume, NULL)) ){
        // ext2 and vfat might get detected from HAL, but the check is
        // incorect and might fail, so let's not believe it for now
        // also do not use it for cdrom fs
        if ( strcmp( fstype, "ext2" ) && strcmp( fstype, "vfat" ) &&
             strcmp( fstype, "fat" ) && strcmp( fstype, "iso9660" ) &&
             strcmp( fstype, "udf" ) )
	   snprintf( subfs_type, MOUNT_LENGTH, "fs=%s,sync,", fstype );
     };
//     options = hal_volume_policy_get_mount_options (drive, volume, NULL);
    }

    mkdir (mount_point, 0775);
    snprintf( cmd, MOUNT_LENGTH, "/bin/mount -t subfs -o %snosuid,nodev,exec %s \"%s\"", subfs_type, device_file, mount_point );

    syslog (LOG_INFO, "registered at resmgrd and called(%d) %s ", system(cmd), cmd);
  }

  // unmount device
  if (udi_to_remove){
    // hald did already the unmount 
    if ( ! rmdir( mount_point ) ){
       syslog (LOG_INFO, "Mount point %s got removed.",
               mount_point);
    }else{
       syslog (LOG_INFO, "Mount point %s can't get removed!",
               mount_point );
    }
  }

  hal_shutdown (hal_context);
 
  return 0;
}

